/*
 * Segment 2 (KEx), SWE LMS-TIC3
 * Copyright (C) 2016 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lmstic3.empfang.erzeugung;

import java.io.File;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.LinkedBlockingQueue;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.InvalidationListener;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.lmstic3.empfang.dekoder.Meldungsdekodierer;
import de.bsvrz.kex.lmstic3.vew.LMSMeldungenListener;
import de.bsvrz.kex.lmstic3.vew.LMSParameter;
import de.bsvrz.kex.lmstic3.vew.LMSTIC3;
import de.bsvrz.kex.lmstic3.vew.LMSTIC3LogTools;
import de.bsvrz.kex.lmstic3.vew.LmsMsg;
import de.bsvrz.kex.lmstic3.vew.TMCMeldung;
import de.bsvrz.kex.lmstic3.vew.daten.ExterneMeldung;
import de.bsvrz.kex.lmstic3.vew.daten.TMCKonvertierungException;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3Converter;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3Exception;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCException;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.dynobj.DynObjektException;
import de.bsvrz.sys.funclib.dynobj.DynamischeObjekte;

/**
 * Modul Meldungserzeugung.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: Meldungserzeugung.java 66499 2017-06-21 12:23:02Z gieseler $
 */

public class Meldungserzeugung extends Thread implements ClientReceiverInterface, ClientSenderInterface,
		InvalidationListener, LMSMeldungenListener, MutableSetChangeListener {

	/**
	 * Worker-Thread.
	 */
	private class MeldungenWorker extends Thread {

		private boolean run = true;

		MeldungenWorker() {
			this.start();
		}

		@Override
		public void run() {
			while (run) {
				try {
					final List<ExterneMeldung> neueMeldungen = meldungsAuftraege.take();
					neueMeldungenVerarbeiten(neueMeldungen);
				} catch (final InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}

		}

		public void stopp() {
			run = false;
		}

		private void neueMeldungenVerarbeiten(final List<ExterneMeldung> neueMeldungen) {
			final List<String> neueTic3Meldungen = new ArrayList<>();
			final List<String> neueTmcMeldungen = new ArrayList<>();

			for (ExterneMeldung m : neueMeldungen) {
				try {
					neueTic3Meldungen.add(createOrUpdateTIC3Meldung(m));
				} catch (final TMCException e) {
					debug.error("Zur Meldung '" + m.getID() + "' konnte keine TIC3-Meldung angelegt werden: "
							+ e.getMessage());
				}

				try {
					final String tmcMeldung = createOrUpdateTMCMeldung(m);
					if (tmcMeldung != null) {
						neueTmcMeldungen.add(tmcMeldung);
					}
				} catch (final TMCException e) {
					debug.error("Zur Meldung '" + m.getID() + "' konnte keine TMC-Meldung angelegt werden: "
							+ e.getMessage());
				}
			}

			// Mengen synchronisieren
			final List<SystemObject> loeschen = new ArrayList<>();

			for (String m : meldungenTIC3.keySet()) {
				if (!neueTic3Meldungen.contains(m)) {
					debug.info("Die nicht mehr bestehende Meldung '" + m + "' wird gelöscht");
					loeschen.add(meldungenTIC3.get(m).getSystemObject());
				}
			}

			meldungenLoeschen(loeschen, meldungsmengeTIC3);

			for (SystemObject s : loeschen) {
				if (meldungenTIC3.containsKey(s.getPid())) {
					meldungenTIC3.remove(s.getPid());
				}
			}

			loeschen.clear();
			for (String m : meldungenTMC.keySet()) {
				if (!neueTmcMeldungen.contains(m)) {
					debug.info("Die nicht mehr bestehende Meldung '" + m + "' wird gelöscht");
					loeschen.add(meldungenTMC.get(m).getSystemObject());
				}
			}

			meldungenLoeschen(loeschen, meldungsmengeTMC);

			for (SystemObject s : loeschen) {
				if (meldungenTMC.containsKey(s.getPid())) {
					meldungenTMC.remove(s.getPid());
				}
			}
		}
	}

	private String createOrUpdateTIC3Meldung(final ExterneMeldung meldung) throws TMCException {
		final String pid = meldung.bildePid();

		ExterneMeldung found = null;

		synchronized (meldungenTIC3) {
			found = meldungenTIC3.get(pid);
		}

		if (found == null) {
			try {
				debug.info("Eine neue TIC3-Meldung '" + meldung.bildePid() + "' wird erzeugt");
				final SystemObject systemObject = vdo.erzeugeObjektInMenge(objektTypExterneMeldungen, pid, pid,
						meldungsmengeTIC3);
				meldung.setSystemObject(systemObject);
				found = meldung;
			} catch (final DynObjektException de) {
				throw new TMCException(de.getMessage());
			}
		} else {
			debug.info("Die Daten der TIC3-Meldung '" + pid + "' werden aktualisiert");
			found.setMeldungsDaten(meldung.getMeldungsDaten());
		}

		try {
			found.anmeldenDav(dav);
			found.publiziereDav(dav, lmsParameter);
			synchronized (meldungenTIC3) {
				meldungenTIC3.put(pid, found);
			}
			return pid;
		} catch (final TIC3Exception e) {
			throw new TMCException(e.getMessage());
		}
	}

	private void meldungenLoeschen(final List<SystemObject> loeschen, final MutableSet menge) {
		for (SystemObject o : loeschen) {
			try {
				vdo.entferneObjektAusMenge((DynamicObject) o, menge, true);
			} catch (final DynObjektException e) {
				debug.error("Das dynamische Meldungsobjekt '" + o.getPid() + "' konnte nicht aus der Menge '"
						+ menge.getName() + "' entfernt werden: " + e.getMessage());
			}
		}
	}

	private String createOrUpdateTMCMeldung(final ExterneMeldung meldung) throws TMCException {
		final String pid = TMCMeldung.bildePid(meldung.getID());
		TMCMeldung found = null;

		final AttributeGroup attributeGroup = dav.getDataModel().getAttributeGroup(TMCMeldung.ATG);
		final Data createData = dav.createData(attributeGroup);

		final TIC3Converter conv = new TIC3Converter(meldung.getMeldungsDaten());

		if (!conv.hasTmcDaten()) {
			debug.info("Die Meldung '" + meldung.getID()
					+ "' besitzt keine TMC-Daten. Es wird kein TMC-Meldungsobjekt angelegt.");
			return null;
		}

		try {
			conv.erzeugeDAVDaten(createData);
		} catch (final TMCKonvertierungException e) {
			debug.warning("Die Meldung '" + meldung.getID() + "' konnte nicht in eine TMC-Meldung konvertiert werden: "
					+ e.getMessage());
			return null;
		}

		synchronized (meldungenTMC) {
			found = meldungenTMC.get(pid);
		}

		if (found == null) {
			try {
				debug.info("Eine neue TMC-Meldung '" + pid + "' wird erzeugt");
				final SystemObject systemObject = vdo.erzeugeObjektInMenge(objektTypTMC, pid, pid, meldungsmengeTMC);
				found = new TMCMeldung(systemObject);
			} catch (final DynObjektException de) {
				throw new TMCException(de.getMessage());
			}
		} else {
			debug.info("Die Daten der TMC-Meldung '" + pid + "' werden aktualisiert");
		}

		try {
			found.setData(createData);
			found.anmeldenDavEmpfang(dav);
			found.publiziereDav(dav);
			synchronized (meldungenTMC) {
				meldungenTMC.put(pid, found);
			}
			return pid;
		} catch (final TIC3Exception e) {
			throw new TMCException(e.getMessage());
		}
	}

	/**
	 * Listener f&uuml;r &Auml;nderungen der Menge der RDS-Meldungen des
	 * Landesmeldestelle.
	 */
	class MengeListener implements MutableSetChangeListener {

		@Override
		public void update(final MutableSet set, final SystemObject[] addedObjects,
				final SystemObject[] removedObjects) {
			LMSTIC3LogTools.log(debug, LmsMsg.FineMengenAenderung);

			for (final SystemObject so : addedObjects) {
				LMSTIC3LogTools.log(debug, LmsMsg.InfoMeldungNeu, so.getPid());
				hinzufuegenMeldung(so);
			}

			for (final SystemObject so : removedObjects) {
				LMSTIC3LogTools.log(debug, LmsMsg.InfoMeldungEntfernt, so.getPid());
				entfernenMeldung(so);
			}
		}
	}

	/**
	 * Attributgruppe der Kommunikationsparameter.
	 */
	private static final String ATTRIBUTGRUPPE_PARAMETER = "atg.lmsKommunikationsParameter";

	/**
	 * Logger f&uuml;r Debug-Ausgaben.
	 */
	private final Debug debug = Debug.getLogger();

	/**
	 * DaV Interface.
	 */
	private ClientDavInterface dav;

	/**
	 * Datenmodell der Konfiguration.
	 */
	private DataModel config;

	// private final String DAV_TYP = "typ.landesMeldeStelle";

	/**
	 * Objektreferenz auf das Objekt vom Typ "typ.tic3LandesMeldeStelle".
	 */
	private ConfigurationObject lms;

	/**
	 * Die Liste der zuletzt empfangenen TMC-Meldungen.
	 */
	// private Map<String, TIC3Meldung> meldungen = new HashMap<>();

	/**
	 * Die Liste der aktuellen TIC3-Meldungen.
	 */
	private Map<String, ExterneMeldung> meldungenTIC3 = new HashMap<>();

	/**
	 * Die Liste der aktuellen TMC-Meldungen.
	 */
	private Map<String, TMCMeldung> meldungenTMC = new HashMap<>();

	/**
	 * Die Menge der TMC-Meldungen der Landesmeldestelle.
	 */
	private MutableSet meldungsmengeTMC;

	/**
	 * Die Menge der TIC3-Meldungen der Landesmeldestelle.
	 */
	private MutableSet meldungsmengeTIC3;

	/**
	 * Die aktuellen Parameter.
	 */
	private final LMSParameter lmsParameter = new LMSParameter();

	/** Datenbeschreibung, mit der Parameter empfangen werden. */
	private DataDescription paramEmpfang;

	/** Meldungsdekodierer. */
	private Meldungsdekodierer meldungsdekodierer;

	/**
	 * Run-Flag.
	 */
	private boolean run = true;

	/** Klasse zum Verwalten dynamischer Objekte. */
	private DynamischeObjekte vdo;

	/** Objekttyp der TMC-Meldungen. */
	private DynamicObjectType objektTypTMC;

	/** Objekttyp der TIC3-Meldungen. */
	private DynamicObjectType objektTypExterneMeldungen;

	private boolean mitValidierung = true;

	private boolean parameterEmpfangen;

	/**
	 * Enth&auml;lt die empfangenen Meldungen.
	 */
	private LinkedBlockingQueue<List<ExterneMeldung>> meldungsAuftraege = new LinkedBlockingQueue<>();

	private MeldungenWorker worker;

	private boolean datenAngemeldet;

	/**
	 * Konstruktor.
	 *
	 * @param dav
	 *            Datenverteiler-Verbindung
	 * @param lms
	 *            Landesmeldestelle
	 * @param mitValidierung
	 *            Schame-Validierung ausf&uuml;hren oder nicht
	 */
	public Meldungserzeugung(final ClientDavInterface dav, final ConfigurationObject lms,
			final boolean mitValidierung) {
		super();
		this.dav = dav;
		this.lms = lms;
		this.mitValidierung = mitValidierung;
		config = this.dav.getDataModel();
		this.setName(this.getClass().getSimpleName());
		vdo = DynamischeObjekte.getInstanz(dav);
		objektTypTMC = (DynamicObjectType) dav.getDataModel().getType(TMCMeldung.TYP);
		if (objektTypTMC == null) {
			throw new RuntimeException("Der Typ '" + TMCMeldung.TYP + "' existiert nicht!");
		}

		objektTypExterneMeldungen = (DynamicObjectType) dav.getDataModel().getType("typ.tmcTic3infoMeldung");
		if (objektTypExterneMeldungen == null) {
			throw new RuntimeException("Der Typ '" + "typ.tmcTic3infoMeldung" + "' existiert nicht!");
		}

		// selbst starten
		start();
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		/**
		 * nichts zu tun
		 */
	}

	/**
	 * Entfernt eine Meldung aus der Meldungsliste.
	 *
	 * @param so
	 *            Systemobjekt der Meldung
	 */
	synchronized void entfernenMeldung(final SystemObject so) {
		synchronized (meldungenTMC) {
			final TMCMeldung tmcMeldung = meldungenTMC.get(so.getPid());
			if (tmcMeldung != null) {
				try {
					tmcMeldung.abmeldenDavEmpfang(dav);
				} catch (final TIC3Exception e) {
					debug.error("abmeldenDavEmpfang: " + e.getMessage());
				}
				meldungenTMC.remove(so.getPid());
			}
		}

		synchronized (meldungenTIC3) {
			final ExterneMeldung externeMeldung = meldungenTIC3.get(so.getPid());
			if (externeMeldung != null) {
				try {
					externeMeldung.abmeldenDavEmpfang(dav);
				} catch (final TIC3Exception e) {
					debug.error("abmeldenDavEmpfang: " + e.getMessage());
				}
				meldungenTIC3.remove(so.getPid());
			}
		}

	}

	/**
	 * F&uuml;gt eine Meldung zur Meldungsliste hinzu.
	 *
	 * @param objekt
	 *            Systemobjekt der Meldung
	 *
	 */
	public void hinzufuegenMeldung(final SystemObject objekt) {
		// synchronized (meldungen) {
		// final TIC3Meldung meldung = new TIC3Meldung(objekt);
		// try {
		// meldung.aktualisierenDav(dav);
		// } catch (final TIC3Exception e) {
		// // das ist hier kein Fehler:
		// // Meldungen von RdcTmcApp, die erst in die Menge eingetragen
		// // und dann mit Daten gef�ttert werden erzeugen sonst hier
		// // einen Fehler
		// // LogTools.log(debug, LmsMsg.ErrorAppException,
		// // e.getMessage());
		// }
		// try {
		// meldung.anmeldenDavEmpfang(dav);
		// } catch (final TIC3Exception e) {
		// LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
		// }
		// meldungen.put(meldung.getPID(), meldung);
		// }
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		// debug.finest("isRequestSupported");
		return false;
	}

	@Override
	public void run() {
		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
			@Override
			public void uncaughtException(final Thread t, final Throwable e) {
				LMSTIC3LogTools.log(debug, LmsMsg.ErrorThreadBeendet, "Meldungserzeugung (" + e + ")"); //$NON-NLS-1$
				run = false;
				if ((meldungsdekodierer != null) && meldungsdekodierer.isAlive()) {
					meldungsdekodierer.stopp();
					try {
						meldungsdekodierer.join();
					} catch (final InterruptedException e1) {
						// e1.printStackTrace();
					}
				}
			}
		});

		meldungenLoeschen();

		synchronized (this) {
			try {
				subscribeData();
				parameterEmpfangen = false;

				while (run) {
					// auf Parameter warten
					if (!parameterEmpfangen) {
						LMSTIC3LogTools.log(debug, LmsMsg.InfoWarteParameter);

						try {
							wait();
						} catch (final InterruptedException e1) {
							debug.warning("unterbrochen");
						}
					}
					parameterEmpfangen = false;

					if (!isEmpfangAktiv()) {
						LMSTIC3LogTools.log(debug, LmsMsg.InfoKeinEmpfang);
						continue;
					}

					if (!parameterGueltig()) {
						continue;
					}

					worker = new MeldungenWorker();

					meldungsdekodierer = new Meldungsdekodierer(lms.getPid(), lmsParameter, mitValidierung);
					meldungsdekodierer.addListener(this);
					while (run) {
						try {
							wait(10000);

							if (!meldungsdekodierer.isAlive()) {
								LMSTIC3LogTools.log(debug, LmsMsg.ErrorThreadBeendet, "Meldungsdekodierer");
								run = false;
								break;
							}

							if (parameterEmpfangen) {
								break;
							}
						} catch (final InterruptedException e) {
							debug.warning("unterbrochen");
						}
					}
					if (meldungsdekodierer != null) {
						meldungsdekodierer.stopp();
					}

					if (worker != null) {
						worker.stopp();
					}
				}
			} catch (final TIC3Exception e) {
				LMSTIC3LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
			}
		}
	}

	private boolean parameterGueltig() {
		final File dir = new File(lmsParameter.getEingangsverzeichnis());

		if (!dir.exists()) {
			debug.warning("Das parametrierte Eingangsverzeichnis '" + lmsParameter.getEingangsverzeichnis()
					+ "' exitiert nicht");
			return false;
		}

		if (!dir.isDirectory()) {
			debug.warning("Das parametrierte Eingangsverzeichnis '" + lmsParameter.getEingangsverzeichnis()
					+ "' ist kein Verzeichnis");
			return false;
		}

		return true;
	}

	private void meldungenLoeschen() {
		loescheMeldungen(LMSTIC3.MENGE_TMC_MELDUNGEN, LMSTIC3.TYP_TMC_MELDUNGEN, "TMC-Meldungen", true);
		loescheMeldungen(LMSTIC3.MENGE_TIC3_MELDUNGEN, LMSTIC3.TYP_TIC3_MELDUNGEN, "TIC3-Meldungen", false);
	}

	/**
	 * Stoppt den Thread.
	 */
	public void stopp() {
		this.run = false;
		this.interrupt();
	}

	/**
	 * Anmeldung aller Daten am DAV.
	 *
	 * @throws TIC3Exception
	 *             bei Ausnahmen
	 */
	protected void subscribeData() throws TIC3Exception {
		final DataDescription dataDescription = null;
		SystemObject[] objekte = null;
		AttributeGroup attributeGroup = null;
		Aspect aspect = null;

		if (datenAngemeldet) {
			return;
		}

		// Anmeldungen zum Senden und Empfangen.
		try {
			// Menge TMC-Meldungen auslesen
			meldungsmengeTMC = lms.getMutableSet("TMCMeldungen");
			if (meldungsmengeTMC == null) {
				throw new TIC3Exception(
						MessageFormat.format(LmsMsg.ErrorLMSObjektMenge.toString(), "TMCMeldungen", lms));
			}

			for (final SystemObject meldung : meldungsmengeTMC.getElements()) {
				LMSTIC3LogTools.log(debug, LmsMsg.InfoMeldungInitialisiert, meldung);
				// meldungenTMC.put(meldung.getPid(), new TMCMeldung(meldung));
			}
			meldungsmengeTMC.addChangeListener(this);

			// Menge TIC3-Meldungen auslesen
			meldungsmengeTIC3 = lms.getMutableSet("TMCTIC3Meldungen");
			if (meldungsmengeTIC3 == null) {
				throw new TIC3Exception(
						MessageFormat.format(LmsMsg.ErrorLMSObjektMenge.toString(), "TMCTIC3Meldungen", lms));
			}

			for (final SystemObject meldung : meldungsmengeTIC3.getElements()) {
				LMSTIC3LogTools.log(debug, LmsMsg.InfoMeldungInitialisiert, meldung);
				// meldungenTIC3.put(meldung.getPid(), new
				// TIC3Meldung(meldung));
			}
			meldungsmengeTIC3.addChangeListener(this);

			//
			// Anmeldungen zum Empfang
			//

			// Kommunikationsparameter
			objekte = new SystemObject[] { lms };
			attributeGroup = config.getAttributeGroup(Meldungserzeugung.ATTRIBUTGRUPPE_PARAMETER);
			aspect = config.getAspect("asp.parameterSoll");
			paramEmpfang = new DataDescription(attributeGroup, aspect);
			debug.finer("Anmeldung zum Empfang von Daten beim DaV als SENKE :" + "(Objekt: " + lms.getName()
					+ ") (ATG: " + paramEmpfang.getAttributeGroup().getName() + ") (ASP: "
					+ paramEmpfang.getAspect().getName() + ") (OT: " + lms.getType().getName() + ")");

			dav.subscribeReceiver(this, objekte, paramEmpfang, new ReceiveOptions(false), // keine
					// Deltas,
					// keine
					// nachgelieferten Daten
					ReceiverRole.receiver());

			// InvalidationListener
			objektTypTMC.addInvalidationListener(this);
			objektTypExterneMeldungen.addInvalidationListener(this);
			datenAngemeldet = true;
		} catch (final DataNotSubscribedException e) {
			// TBD: Die Exception ist wiedersprüchlich dokumentiert --> Roland
			LMSTIC3LogTools.log(debug, LmsMsg.ErrorAnmeldungEmpfang, dataDescription);
			throw new RuntimeException(e);
		}
	}

	@Override
	public void update(final ResultData[] results) {
		for (final ResultData datensatz : results) {
			if (datensatz.hasData()) {
				if (datensatz.getDataDescription().equals(paramEmpfang)) {
					debug.info("Parameterdatensatz erhalten für ",
							datensatz.getObject() + ", " + datensatz.getDataDescription().getAttributeGroup());
					lmsParameter.aktualisieren(datensatz.getData());
					synchronized (this) {
						parameterEmpfangen = true;
						this.notify();
					}
				}
			}
		}
	}

	@Override
	public void invalidObject(final DynamicObject dynamicObject) {
		debug.info("Meldung: " + dynamicObject.getPid() + " wurde invalidiert!");
		entfernenMeldung(dynamicObject);
	}

	@Override
	public void empfangen(final List<ExterneMeldung> externemeldungen) {
		meldungsAuftraege.add(externemeldungen);
	}

	private void handleTMCSetUpdate(final SystemObject[] addedObjects, final SystemObject[] removedObjects) {
		synchronized (meldungenTMC) {
			for (final SystemObject so : addedObjects) {
				debug.info("Die neue TMC-Meldung '" + so.getPid() + "' wurde der Menge '" + "TMCMeldungen"
						+ "' hinzugefügt!");
				// meldungenTMC.put(so.getPid(), new TMCMeldung(so));
			}

			// for (final SystemObject so : removedObjects) {
			// debug.info("Die TMC-Meldung '" + so.getPid() + "' wird
			// entfernt!");
			// meldungenTMC.remove(so.getPid());
			// }
		}
	}

	private void handleTIC3SetUpdate(final SystemObject[] addedObjects, final SystemObject[] removedObjects) {
		synchronized (meldungenTIC3) {
			for (final SystemObject so : addedObjects) {
				debug.info("Die neue TMC-Meldung '" + so.getPid() + "' wurde der Menge '" + "TMCTIC3Meldungen"
						+ "' hinzugefügt!");
			}
		}
	}

	private class MeldungListenerSync implements MutableSetChangeListener {
		private Object lock;
		private ArrayList<SystemObject> mliste;

		MeldungListenerSync(final Object myLock, final ArrayList<SystemObject> arrayList) {
			lock = myLock;
			mliste = arrayList;
		}

		@Override
		public void update(final MutableSet set, final SystemObject[] addedObjects,
				final SystemObject[] removedObjects) {
			final StringBuffer output = new StringBuffer();

			if (removedObjects != null && removedObjects.length > 0) {
				output.append("gelöschte Objekte: ");
				for (final SystemObject so : removedObjects) {
					mliste.remove(so);
					output.append(so.getPid() + " ");
					try {
						so.invalidate();
					} catch (final ConfigurationChangeException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}

				debug.info(output.toString());
			}

			if ((mliste.size() == 0) && (lock != null)) {
				synchronized (lock) {
					lock.notify();
				}
			}
		}
	}

	private void loescheMeldungen(final String menge, final String typ, final String name,
			final boolean mitParamCheck) {

		try {
			final Object myLock = new Object();
			final MutableSet meldungsmenge = lms.getMutableSet(menge);
			final List<SystemObject> loeschen = new ArrayList<>();

			if (mitParamCheck) {
				for (SystemObject element : meldungsmenge.getElements()) {
					if (isMeldungZuLoeschen(element)) {
						loeschen.add(element);
					}
				}
			} else {
				loeschen.addAll(meldungsmenge.getElements());
			}

			debug.info(name + " am Objekt '" + lms + "' werden gelöscht!");
			meldungsmenge.addChangeListener(new MeldungListenerSync(myLock, new ArrayList<>(loeschen)));
			meldungsmenge.remove(loeschen.toArray(new SystemObject[loeschen.size()]));
			if (loeschen.size() != 0) {
				final List<SystemObject> mi = new ArrayList<>(loeschen);
				for (final SystemObject object : mi) {
					debug.info("Meldung '" + object + "' wird gelöscht!");
					meldungsmenge.remove(object);
				}

				synchronized (myLock) {
					try {
						myLock.wait(10000);
					} catch (final InterruptedException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			} else {
				debug.info("keine Meldungen aus der Menge '" + menge + "' zu löschen");
			}

			// Meldungen, die nicht in der Menge enthalten sind:
			final List<SystemObject> robjekte = dav.getDataModel().getType(typ).getElements();
			if (robjekte.size() > 0) {
				debug.info("lösche weitere " + robjekte.size() + " " + name);
				int i = 0;
				for (final SystemObject so : robjekte) {
					boolean del = true;
					if (mitParamCheck && !isMeldungZuLoeschen(so)) {
						del = false;
					}

					if (del) {
						debug.info("Meldung '" + so + "' wird gelöscht!");
						so.invalidate();
						i++;
					}
				}
				if (i > 0) {
					debug.info("weitere " + i + " " + name + " wurden gelöscht");
				}
			}

		} catch (final Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Pr&uuml;ft, ob die Meldung zu l&ouml;schen ist.
	 * <p>
	 * Es werden alle Meldungen gel&ouml;scht, die keine Parameter besitzen.
	 * </p>
	 * 
	 * @param so
	 *            {@link SystemObject}
	 * @return <code>true</code>, wenn zu l&ouml;schen, sonst <code>false</code>
	 */
	private boolean isMeldungZuLoeschen(final SystemObject so) {
		final AttributeGroup attributeGroup = dav.getDataModel().getAttributeGroup(TMCMeldung.ATG);
		final Aspect aspect = dav.getDataModel().getAspect("asp.parameterSoll");
		final DataDescription dataDescription = new DataDescription(attributeGroup, aspect);
		final ResultData data = dav.getData(so, dataDescription, 0);

		if (data.hasData()) {
			return false;
		}

		return true;
	}

	@Override
	public void update(final MutableSet set, final SystemObject[] addedObjects, final SystemObject[] removedObjects) {
		if (set.equals(meldungsmengeTMC)) {
			handleTMCSetUpdate(addedObjects, removedObjects);
		}

		if (set.equals(meldungsmengeTIC3)) {
			handleTIC3SetUpdate(addedObjects, removedObjects);
		}
	}

	private boolean isEmpfangAktiv() {
		return !lmsParameter.getEingangsverzeichnis().isEmpty();
	}
}
