/*
 * Segment 2 (KEx), SWE LMS-TIC3
 * Copyright (C) 2016 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lmstic3.vew;

import java.io.File;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import com.bitctrl.resource.Configuration;
import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.lmstic3.empfang.dekoder.TIC3InfoDatei;
import de.bsvrz.kex.lmstic3.empfang.dekoder.TIC3InfoDateiDaten;
import de.bsvrz.kex.lmstic3.empfang.dekoder.TIC3InfoDateiException;
import de.bsvrz.kex.lmstic3.vew.daten.DavDatenModell;
import de.bsvrz.kex.lmstic3.vew.daten.ExterneMeldung;
import de.bsvrz.kex.lmstic3.vew.daten.TMCKonvertierungException;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3Converter;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3Exception;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3Meldung;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3MeldungsDaten;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCException;
import de.bsvrz.sys.funclib.dynobj.DynObjektException;
import de.bsvrz.sys.funclib.dynobj.DynamischeObjekte;

/**
 * Liest die Datei 'Configuration.VersandMeldung' ein, erzeugt daraus eine eine
 * TMC-Meldung im Datenverteiler, versendet die Daten unter dem Aspekt
 * 'tmcSenden'. Die Ablaufzeit der Meldung wird auf die aktuelle Zeit +
 * Configuration.Gueltigkeitsdauer gesetzt. Das Programm bleibt bis zum Ende der
 * Ablaufzeit als Quelle angemeldet.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: $
 */
public class VersendeMeldung implements ClientSenderInterface {
	/** Der Logger der Klassse. */
	private static Logger log = Logger.getLogger(VersendeMeldung.class.getName());

	private static ClientDavConnection verbindung;

	private static Configuration konfig;

	/** Klasse zum Verwalten dynamischer Objekte. */
	private static DynamischeObjekte vdo;

	private static List<DynamicObject> erzeugteObjekte = new ArrayList<>();

	@BeforeClass
	public static void beforeClass() throws Exception {
		// final ObjektFactory factory;

		final Level level;

		VersendeMeldung.konfig = Configuration.getConfiguration();
		level = Level.parse(VersendeMeldung.konfig.getString("debug.level"));
		LoggerTools.setLoggerLevel(level);
		VersendeMeldung.log.setLevel(level);

		VersendeMeldung.log.info("Stelle Verbindung zum Datenverteiler her ...");
		VersendeMeldung.verbindung = new ClientDavConnection();
		VersendeMeldung.verbindung.getClientDavParameters()
				.setDavCommunicationAddress(VersendeMeldung.konfig.getString("dav.host"));
		VersendeMeldung.verbindung.getClientDavParameters()
				.setDavCommunicationSubAddress(VersendeMeldung.konfig.getInt("dav.port"));
		VersendeMeldung.verbindung.connect();
		VersendeMeldung.verbindung.login(VersendeMeldung.konfig.getString("dav.benutzer"),
				VersendeMeldung.konfig.getString("dav.kennwort"));
	}

	@AfterClass
	public static void afterClass() throws Exception {
		for (DynamicObject objekt : erzeugteObjekte) {
			vdo.entferneObjektAusMenge(objekt, meldungsmengeTMC, true);
		}
	}

	/**
	 * Landesmeldestelle.
	 */
	private final String landesmeldeStelle = "lms.test";

	private static MutableSet meldungsmengeTMC;

	private List<ExterneMeldung> bildeMeldungenAusDatei(final String dateiname)
			throws TIC3InfoDateiException, TIC3Exception {
		final TIC3InfoDatei tid = new TIC3InfoDatei(new File(dateiname));
		final List<ExterneMeldung> dateimeldungen = new ArrayList<>();

		final TIC3InfoDateiDaten bildeMeldungen = tid.bildeMeldungen(false);

		for (TIC3MeldungsDaten daten : bildeMeldungen.getMeldungen()) {
			dateimeldungen.add(new TIC3Meldung(daten));
		}

		return dateimeldungen;
	}

	private List<ExterneMeldung> bildeMeldungenAusDateien() throws TIC3InfoDateiException, TIC3Exception {
		final String dateiliste = VersendeMeldung.konfig.getString("VersandMeldung");
		final List<ExterneMeldung> dateimeldungen = new ArrayList<>();

		for (final String datei : dateiliste.split(",")) {
			dateimeldungen.addAll(bildeMeldungenAusDatei(datei));
		}

		return dateimeldungen;
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		// TODO Auto-generated method stub

	}

	/**
	 * Erzeugt TMC-Meldungen und versendet diese an den DaV.
	 */
	@Test
	public void erzeugeMeldungen() {

		vdo = DynamischeObjekte.getInstanz(VersendeMeldung.verbindung);

		final SystemObject lms = VersendeMeldung.verbindung.getDataModel().getObject(landesmeldeStelle);

		try {
			DavDatenModell.getInstanz().setVerbindung(VersendeMeldung.verbindung);

			meldungsmengeTMC = ((ConfigurationObject) lms).getMutableSet("TMCMeldungen");

			final GregorianCalendar cal = new GregorianCalendar();

			boolean found = false;

			for (final ExterneMeldung meldung : bildeMeldungenAusDateien()) {

				System.out.println("Sende Daten für die Meldung '" + meldung + "'");

				found = true;

				createOrUpdateTMCMeldung(meldung);

				final long dauer = Long.parseLong(VersendeMeldung.konfig.getString("Gueltigkeitsdauer"));

				cal.setTimeInMillis(System.currentTimeMillis() + dauer);
			}

			// bis zum Ende der Gültigkeit warten, damit Quelle nicht
			// verschwindet
			if (found) {
				Thread.sleep(cal.getTimeInMillis() - System.currentTimeMillis());
			}
		} catch (final Exception e) {
			e.printStackTrace();
			Assert.fail(e.getMessage());
		}
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		// TODO Auto-generated method stub
		return false;
	}

	private TIC3Meldung neueMeldung(final TIC3MeldungsDaten meldungsdaten) throws TIC3Exception {
		final ConfigurationArea area = VersendeMeldung.verbindung.getLocalConfigurationAuthority()
				.getConfigurationArea();

		final String pid = meldungsdaten.bildePid();
		SystemObject so = VersendeMeldung.verbindung.getDataModel().getObject(pid);
		if (so == null) {
			try {
				so = area.createDynamicObject(
						(DynamicObjectType) VersendeMeldung.verbindung.getDataModel().getType("typ.tmcVerkehrsMeldung"),
						pid, pid);
				System.out.println("Die Meldung " + pid + " wurde angelegt");
			} catch (final ConfigurationChangeException e) {
				throw new TIC3Exception(e.getMessage());
			}
		} else {
			System.out.println("Die Meldung " + pid + " existiert bereits");
		}

		final TIC3Meldung rdsmeldung = new TIC3Meldung(so, meldungsdaten);

		return rdsmeldung;
	}

	private String createOrUpdateTMCMeldung(final ExterneMeldung meldung) throws TMCException {
		final String pid = TMCMeldung.bildePid(meldung.getID());

		// TMCMeldung found = meldungenTMC.get(pid);
		TMCMeldung found = null;

		final DynamicObjectType objektTypTMC = (DynamicObjectType) VersendeMeldung.verbindung.getDataModel()
				.getType(TMCMeldung.TYP);

		if (found == null) {
			try {
				final SystemObject systemObject = vdo.erzeugeObjektInMenge(objektTypTMC, pid, pid, meldungsmengeTMC);
				System.out.println("Eine neue TMC-Meldung '" + pid + "' wird erzeugt");
				found = new TMCMeldung(systemObject);
				erzeugteObjekte.add((DynamicObject) systemObject);
			} catch (final DynObjektException de) {
				throw new TMCException(de.getMessage());
			}
			// } else {
			// System.out.println("Die Daten der TMC-Meldung '" + pid + "'
			// werden aktualisiert");
		}

		final AttributeGroup attributeGroup = VersendeMeldung.verbindung.getDataModel()
				.getAttributeGroup(TMCMeldung.ATG);
		final Aspect aspect = VersendeMeldung.verbindung.getDataModel().getAspect("asp.tmcSenden");

		final DataDescription dataDescription = new DataDescription(attributeGroup, aspect);

		final Data createData = VersendeMeldung.verbindung.createData(attributeGroup);

		final TIC3Converter conv = new TIC3Converter(meldung.getMeldungsDaten());
		try {
			conv.erzeugeDAVDaten(createData);
			setAblaufzeit(createData);
			found.setData(createData);
		} catch (final TMCKonvertierungException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		final SystemObject[] objekte = new SystemObject[] { found.getSystemObject() };

		try {
			VersendeMeldung.verbindung.subscribeSender(this, objekte, dataDescription, SenderRole.source());
			final ResultData meldungsdaten = new ResultData(found.getSystemObject(), dataDescription,
					System.currentTimeMillis(), createData);
			try {
				VersendeMeldung.verbindung.sendData(meldungsdaten);
			} catch (final DataNotSubscribedException | SendSubscriptionNotConfirmed e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			// dav.subscribeSender(this, _systemObject, _dataParameter,
			// SenderRole.sender());
		} catch (final OneSubscriptionPerSendData e) {
			e.printStackTrace();
		}

		return found.getPID();
	}

	private void setAblaufzeit(final Data davData) {
		final DateFormat formatter = new SimpleDateFormat();
		final GregorianCalendar cal = new GregorianCalendar();
		final long dauer = Long.parseLong(VersendeMeldung.konfig.getString("Gueltigkeitsdauer"));
		cal.setTimeInMillis(System.currentTimeMillis() + dauer);
		System.out.println("Ablaufzeit:       " + formatter.format(cal.getTime()));
		final Data item = davData.getItem("TMCDaten");

		final Data vi = item.getItem("VerwaltungsInformationen");

		final Array array = vi.getArray("EndeZeit");
		array.setLength(1);
		array.getItem(0).asTimeValue().setMillis(cal.getTimeInMillis());
	}
}
