/*
 * Segment 2 (KEx), SWE LMS-TIC3
 * Copyright (C) 2016 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lmstic3.vew.daten.tic.elemente;

import de.bsvrz.kex.lmstic3.vew.daten.tic.AbstractTIC3InfoElement;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3InfoAttribut;

/**
 * TIC3-Quelle.
 * <p>
 * Entspricht TIC3-Typ 'UnspecificVersionReferenceExtendedType'.
 * </p>
 *
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: TIC3Quelle.java 63650 2016-09-22 17:11:32Z gieseler $
 */
public class TIC3Quelle extends AbstractTIC3InfoElement {
	/**
	 * TIC3 Info XML Element: Source (SRC), A list of references to and information about the source data used to create the data contained in this object.
	 */
	public static final String XML_ELEMENT_QUELLE = "Source";

	/**
	 * Beschreibung.
	 * <p>Beschreibung aus Wert des 'description'-Attributes.</p>
	 */
	private TIC3InfoAttribut beschreibung = new TIC3InfoAttribut("description", "Beschreibung");
	
	/**
	 * Objekttyp.
	 * <p>TIC3 Info XML Element: ObjectType (OTY), The Identifier of the object type.</p>
	 */
	private final TIC3Text typ = new TIC3Text("ObjectType", "ObjektTyp");
	

	/**
	 * TicId.
	 * <p>TIC3 Info XML Element: TicId (TID), An identifier which is unique across all systems using TIC for this object type.</p>
	 */
	private final TIC3Text tic3Id = new TIC3Text("TicId", "TicId");

	/**
	 * TIC3 Datenerzeugungszeit.
	 * <p>TIC Info XML Element: CreateTime (CRT), The date and time when the data contained in this object version was created.</p>
	 */
	private final TIC3Zeitstempel datenErzeugungsZeit = new TIC3Zeitstempel("CreateTime", "TIC3DatenErzeugungsZeit");

	/**
	 * Quelldatenproduzent.
	 * <p>TIC3 Info XML Element: SourceDataProducer (SDP), The data producer of the source data. A reference to "Organization".</p>
	 */
	private final TIC3Text quelldatenproduzent = new TIC3Text("SourceDataProducer", "TIC3Quelldatenproduzent");
	
	/**
	 * Quelldaten-Beschreibung.
	 * <p>TIC3 Info XML Element: SourceDescription (SDS), A culture-dependent description of the source data.</p>
	 */
	private final TIC3Text quelldatenBeschreibung = new TIC3Text("SourceDescription", "TIC3QuelldatenBeschreibung");

	/**
	 * Ergebnis erzeugen..
	 * <p>TIC3 Info XML Element: CreateResult (CRE), Settings used to create the data contained in this object.</p>
	 */
	private final TIC3ErgebnisErzeugung ergebnisErzeugung = new TIC3ErgebnisErzeugung("CreateResult", "TIC3ErgebnisErzeugung");

	/**
	 * Erzeugt ein neues Objekt.
	 */
	public TIC3Quelle() {
		this(XML_ELEMENT_QUELLE, null);
	}
	
	/**
	 * Erzeugt ein neues Objekt.
	 *
	 * @param xmlName
	 *            Name des XML-ELementes
	 * @param davName
	 *            Name des DAV-Attributes
	 */
	public TIC3Quelle(final String xmlName, final String davName) {
		super(xmlName, davName);
		addAttribut(beschreibung);
		addKindElement(datenErzeugungsZeit);
		addKindElement(ergebnisErzeugung);
		addKindElement(quelldatenBeschreibung);
		addKindElement(quelldatenproduzent);
		addKindElement(tic3Id);
		addKindElement(typ);
	}
	
	
	@Override
	public String toString() {
		return this.getClass().getSimpleName() + ": Beschreibung='" + beschreibung.getXmlValue(); 
	}
}
